# -*- coding= utf-8 -*-

import inspect
from pydantic import BaseModel
from redis.asyncio import Redis
from rorm.orm_model import ORMModel
from typing import Awaitable, Callable, Type, TypeVar, get_args

T = TypeVar("T", bound=ORMModel)

_ORMS: dict[str, Type[ORMModel]] = {}


class ORMCache(BaseModel):

    def __init__(self, client: Redis) -> None:
        super().__init__()
        self.__list_copy: dict[str, set[str]] = {}
        self.__dict_copy: dict[str, set[str]] = {}
        self.__client = client

    def unique_id(self) -> str:
        return self.__class__.__qualname__

    @classmethod
    def register(cls, _cls: Type[T]) -> Type[T]:
        if _cls.__qualname__ in _ORMS:
            raise ValueError(
                f"{_cls.__qualname__} has already been registered")
        _ORMS[_cls.__qualname__] = _cls
        return _cls

    async def save_all(self):
        modified: dict[str, ORMModel] = dict()
        del_fields: set[str] = set()
        for attr_name, attr in self.__dict__.items():
            if attr_name.startswith("_ORMCache"):
                continue
            if isinstance(attr, ORMModel):
                if not attr.is_dirty():
                    continue
                modified[
                    f"{attr_name}:{attr.__class__.__qualname__}:{attr.unique_id()}"] = attr
            elif isinstance(attr, list | set):
                list_copy: set[str] = set()
                for item in attr:
                    if not isinstance(item, ORMModel):
                        break
                    list_copy.add(
                        f"{attr_name}:{item.__class__.__qualname__}:{item.unique_id()}")
                    if not item.is_dirty():
                        continue
                    modified[
                        f"{attr_name}:{item.__class__.__qualname__}:{item.unique_id()}"] = item
                diff = self.__list_copy.get(attr_name, set()) - list_copy
                del_fields |= diff
                self.__list_copy[attr_name] = list_copy
            elif isinstance(attr, dict):
                dict_copy: set[str] = set()
                for _, item in attr.items():
                    if not isinstance(item, ORMModel):
                        break
                    dict_copy.add(
                        f"{attr_name}:{item.__class__.__qualname__}:{item.unique_id()}")
                    if not item.is_dirty():
                        continue
                    modified[
                        f"{attr_name}:{item.__class__.__qualname__}:{item.unique_id()}"] = item
                diff = self.__dict_copy.get(attr_name, set()) - dict_copy
                del_fields |= diff
                self.__dict_copy[attr_name] = dict_copy

        if modified:
            result = self.__client.hmset(
                self.unique_id(), mapping={key: value.encode() for key, value in modified.items()})
            if isinstance(result, Awaitable):
                await result
            for _, value in modified.items():
                value.clear_dirty()
        if del_fields:
            result = self.__client.hdel(self.unique_id(), *del_fields)
            if isinstance(result, Awaitable):
                await result

    async def load_from_redis(self, err_callback: Callable[[str, Exception], bool] | None = None) -> bool:
        result: Awaitable[dict[str, bytes]] | dict[str, bytes] = self.__client.hgetall(
            self.unique_id())
        if isinstance(result, Awaitable):
            result = await result
        for key, value in result.items():
            try:
                if isinstance(key, bytes):
                    key = key.decode()
                sp = key.split(":")
                if len(sp) != 3:
                    continue
                attr_name, orm_qualname, unique_id = sp
                if attr_name not in self.__annotations__ or \
                        attr_name not in self.__dict__ or \
                        orm_qualname not in _ORMS:
                    continue

                attr = self.__dict__[attr_name]
                if isinstance(attr, ORMModel):
                    attr = _ORMS[orm_qualname].decode(value)
                    self.__dict__[attr_name] = attr
                elif isinstance(attr, list | set):
                    arg_type = get_args(self.__annotations__[attr_name])[0]
                    if not inspect.isclass(arg_type) or not issubclass(arg_type, ORMModel):
                        continue
                    m = _ORMS[orm_qualname].decode(value)
                    if isinstance(attr, list):
                        attr.append(m)
                    else:
                        attr.add(m)
                    if attr_name not in self.__list_copy:
                        self.__list_copy[attr_name] = set()
                    self.__list_copy[attr_name].add(key)
                elif isinstance(attr, dict):
                    arg_type = get_args(self.__annotations__[attr_name])[1]
                    if not inspect.isclass(arg_type) or not issubclass(arg_type, ORMModel):
                        continue
                    m = _ORMS[orm_qualname].decode(value)
                    attr[unique_id] = m
                    if attr_name not in self.__dict_copy:
                        self.__dict_copy[attr_name] = set()
                    self.__dict_copy[attr_name].add(key)
                else:
                    raise TypeError(
                        f"{self.__class__.__qualname__}:{attr_name} {type(attr)} is not supported")
            except Exception as e:
                if err_callback and not err_callback(key, e):
                    return False
        return True



# -*- coding= utf-8 -*-

import unittest
from pydantic import Field
from redis.asyncio import Redis
from rorm.orm_model import ORMModel
from rorm.orm_cache import ORMCache


@ORMCache.register
class RoleInfo(ORMModel):
    role_id: int = Field(0, title="角色id")
    role_name: str = Field("", title="角色名")
    role_level: int = Field(0, title="角色等级")

    def __init__(self,  /, **data) -> None:
        super().__init__(**data)

    def unique_id(self) -> str:
        return f"{self.role_id}"


@ORMCache.register
class ItemInfo(ORMModel):
    item_id: int = Field(0, title="物品id")
    item_name: str = Field("", title="物品名")
    item_count: int = Field(0, title="物品数量")

    def __init__(self,  /, **data) -> None:
        super().__init__(**data)

    def unique_id(self) -> str:
        return f"{self.item_id}"


@ORMCache.register
class MailInfo(ORMModel):
    mail_id: int = Field(0, title="邮件id")
    mail_title: str = Field("", title="邮件标题")
    mail_content: str = Field("", title="邮件内容")
    attarchments: list[ItemInfo] = []

    def __init__(self,  /, **data) -> None:
        super().__init__(**data)

    def unique_id(self) -> str:
        return f"{self.mail_id}"


class Player(ORMCache):

    role: RoleInfo = RoleInfo()
    items: dict[str, ItemInfo] = {}
    mails: list[MailInfo] = []
    tmp: list[str] = []

    def __init__(self, client: Redis, user_id: int) -> None:
        super().__init__(client)
        self.__user_id = user_id

    def unique_id(self) -> str:
        return f"{self.__class__.__qualname__}:{self.__user_id}"


class TestCache(unittest.IsolatedAsyncioTestCase):

    async def test_save(self):
        client = await Redis.from_url("redis://localhost:6379/0")
        player = Player(client, 10001)
        player.role = RoleInfo(
            role_id=10001, role_name="rorm-py", role_level=1)
        for i in range(1, 6):
            player.items[f"{i}"] = ItemInfo(
                item_id=i, item_name=f"item-{i}", item_count=i)
        for i in range(1, 6):
            player.mails.append(MailInfo(
                mail_id=i,
                mail_title=f"mail-{i}",
                mail_content=f"mail-{i}",
                attarchments=[ItemInfo(
                    item_id=i,
                    item_name=f"item-{i}",
                    item_count=i)]
            ))
        player.tmp.append("not save")
        await player.save_all()
        await client.close()

    async def test_load(self):
        client = await Redis.from_url("redis://localhost:6379/0")
        player = Player(client, 10001)
        await player.load_from_redis()
        print(player.model_dump_json())
        await client.close()

    async def test_modify(self):
        client = await Redis.from_url("redis://localhost:6379/0")
        player = Player(client, 10001)
        await player.load_from_redis()

        player.role.role_level = 2
        for item in player.items.values():
            item.item_count += 1

        for mail in player.mails:
            mail.mail_content = "modify"
            for item in mail.attarchments:
                item.item_count += 1

        await player.save_all()

    async def test_delete(self):
        client = await Redis.from_url("redis://localhost:6379/0")
        player = Player(client, 10001)
        await player.load_from_redis()

        player.role = RoleInfo()
        for key in list(player.items.keys()):
            if int(key) % 2 == 0:
                del player.items[key]

        for mail in list(player.mails):
            if mail.mail_id % 2 == 0:
                player.mails.remove(mail)

        await player.save_all()



from setuptools import setup, find_packages

setup(
    name="rorm-py",
    version="1.0.0",
    author="刘虎",
    author_email="huxiaoheigame@gmail.com",
    description="A simple ORM cache library for redis",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/huxiaohei/rorm-py",  # 项目的主页
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=[
        "annotated-types == 0.7.0",
        "pydantic == 2.8.2",
        "pydantic_core == 2.20.1",
        "redis == 5.0.7",
        "setuptools == 75.1.0",
        "typing_extensions == 4.12.2"
    ],
    license="MIT"
)



# -*- coding= utf-8 -*-

from typing import Any, Self
from pydantic import BaseModel


class ORMModel(BaseModel):

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        self.__copy: Self | None = None

    def unique_id(self) -> str:
        raise NotImplementedError

    def __delattr__(self, _: str) -> None:
        raise AttributeError("ORMModel is immutable")

    def is_dirty(self) -> bool:
        return self != self.__copy

    def encode(self) -> str:
        return self.model_dump_json()

    @classmethod
    def decode(cls, data:  str | bytes | bytearray) -> Self:
        m = cls.model_validate_json(data)
        m.clear_dirty()
        return m

    def clear_dirty(self) -> None:
        self.__copy = self.model_copy(deep=True)
